from masonite.commands import Command


class TenancyDelete(Command):
    """
    Delete all tenants or a specific tenant.

    tenancy:delete
        {--tenants=default : List of tenants to delete}
    """

    def __init__(self, application):
        super().__init__()
        self.app = application
        self.tenancy = self.app.make("multitenancy")

    def get_tenants(self):
        """Returns a list of all tenants."""
        tenants = self.option("tenants")
        try:
            if tenants == "default":
                tenants = self.tenancy.get_tenants()
            else:
                tenants = tenants.split(",")
                tenants = [self.tenancy.get_tenant(tenant) for tenant in tenants]
            return tenants
        except Exception as e:
            self.error(e)
        return []

    def handle(self):
        tenants = self.get_tenants()

        if len(tenants) == 0:
            self.error("No tenants found!")
            exit()

        for tenant in tenants:
            self.app.make("multitenancy").delete(tenant)

        self.info("All tenants deleted!")



from masonite.commands.Command import Command
from masoniteorm.migrations import Migration


class TenancyMigrateRefresh(Command):
    """
    Refreshes database of all tenants or of a specific tenant.

    tenancy:migrate:refresh
        {--tenants=default : List of tenants to migrate}
        {--m|migration=all : Migration's name to be migrated}
        {--d|directory=databases/migrations : The location of the migration directory}
    """

    def __init__(self, application):
        super().__init__()
        self.app = application
        self.tenancy = self.app.make("multitenancy")

    def migration(self, tenant):
        self.tenancy.setup_connection(tenant)

        return Migration(
            command_class=self,
            connection=tenant.database,
            migration_directory=self.option("directory"),
            config_path=None,
            schema=None,
        )

    def handle(self):
        tenants = self.tenancy.get_tenants(self.option("tenants"))

        if len(tenants) == 0:
            self.error("No tenants found!")
            exit()

        for tenant in tenants:
            self.info(f"Refreshing tenant: {tenant.name}")
            self.warning("=====================START=====================")
            self.migration(tenant).refresh(self.option("migration"))
            self.warning("======================END======================")



from masonite.request.request import Request
from .models.Tenant import Tenant
from masonite.configuration import config
from masoniteorm.connections import ConnectionResolver


class MultiTenancy:
    def __init__(self, application):
        self.app = application
        self.tenant_configs = config("multitenancy.tenants")

    def get_tenants(self, tenants):
        """Returns a list of all tenants."""
        try:
            if tenants == "default":
                tenants = self.__all()
            else:
                tenants = tenants.split(",")
                tenants = [self.__one(tenant) for tenant in tenants]
            return tenants
        except Exception as e:
            print(e)
        return []

    def __all(self):
        """Returns a list of all tenants."""
        return Tenant.all()

    def __one(self, tenant_name):
        """Returns a tenant by name."""
        tenant = Tenant.where("database", tenant_name).first()
        if not tenant:
            raise Exception(f"Tenant: `{tenant_name}` not found!")
        return tenant

    def setup_connection(self, tenant):
        self.tenant_configs.update({"default": tenant.database})
        ConnectionResolver().set_connection_details(self.tenant_configs)

    def delete(self, tenant):
        """Deletes a tenant."""
        tenant.delete()

    def get_subdomain(self, request: Request):
        """Returns the subdomain for the current request."""
        hosts = [request.environ.get("HTTP_HOST"), request.environ.get("HTTP_X_FORWARDED_HOST")]
        subdomains = []
        for host in hosts:
            if host:
                items = host.split(".")
                if len(items) > 2:
                    subdomains.append(items[0])
        return subdomains, hosts

    def __reset_connection(self):
        ConnectionResolver().set_connection_details(config("database.databases"))

    def find_tenant(self, request: Request):
        """Finds the tenant for the current request."""
        subdomains, hosts = self.get_subdomain(request)
        subdomains = tuple(subdomains)
        hosts = tuple(hosts)
        print(hosts, subdomains)
        try:
            self.__reset_connection()
            tenant = Tenant.where_raw("database in {database}".format(database=subdomains)).first()
            if tenant:
                self.setup_connection(tenant)
                request.tenant = tenant
        except Exception as e:
            print(e)
        return request



from masonite.middleware import Middleware


class TenantFinderMiddleware(Middleware):
    """Middleware to find the tenant for the current request."""

    def before(self, request, response):
        return request.app.make("multitenancy").find_tenant(request)

    def after(self, request, response):
        """Return the response."""
        return response



"""multitenancy Settings"""

"""
|--------------------------------------------------------------------------
| MultiTenancy
|--------------------------------------------------------------------------
|
| Multitenancy is a feature that allows you to have multiple tenants in your
| application. This is useful for things like a company having multiple
| websites, or a company having multiple apps.
|
"""

TENANTS = {}



from masonite.commands import Command
from ..models.Tenant import Tenant


class TenancyCreate(Command):
    """
    Creates a new tenant.

    tenancy:create
    """

    def __init__(self, application):
        super().__init__()
        self.app = application

    def handle(self):
        name = self.ask("Name: ")
        domain = self.ask("Domain: ")
        database = self.ask("Database: ")

        if not name:
            self.error("Name is required!")
            exit()

        if not domain:
            self.error("Domain is required!")
            exit()

        if not database:
            self.error("Database name is required!")
            exit()

        tenant = Tenant.where("domain", domain).or_where("database", database).first()
        if tenant:
            self.error("Tenant already exists!")
            exit()

        tenant = Tenant.create(name=name, domain=domain, database=database)
        self.info("Tenant created!")



"""CreateTenantsTable Migration."""

from masoniteorm.migrations import Migration


class CreateTenantsTable(Migration):
    def up(self):
        """
        Run the migrations.
        """
        with self.schema.create("tenants") as table:
            table.increments("id")
            table.string("name")
            table.string("domain").unique()
            table.string("database").unique()
            table.boolean("status").default(False)
            table.timestamps()

    def down(self):
        """
        Revert the migrations.
        """
        self.schema.drop("tenants")



from masonite.commands.Command import Command
from masoniteorm.seeds import Seeder
from inflection import camelize, underscore


class TenancySeed(Command):
    """
    Seed data to all tenants or to a specific tenant.

    tenancy:seed:run
        {--tenants=default : List of tenants to run seeder}
        {--dry : If the seed should run in dry mode}
        {table=None : Name of the table to seed}
        {--d|directory=databases/seeds : The location of the seed directory}
    """

    def __init__(self, application):
        super().__init__()
        self.app = application
        self.tenancy = self.app.make("multitenancy")

    def seed(self, tenant):
        self.tenancy.setup_connection(tenant)

        seeder = Seeder(
            dry=self.option("dry"),
            seed_path=self.option("directory"),
            connection=tenant.database,
        )

        if self.argument("table") == "None":
            seeder.run_database_seed()
            seeder_seeded = "Database Seeder"
        else:
            table = self.argument("table")
            seeder_file = f"{underscore(table)}_table_seeder.{camelize(table)}TableSeeder"
            seeder.run_specific_seed(seeder_file)
            seeder_seeded = f"{camelize(table)}TableSeeder"

        self.line(f"<info>{seeder_seeded} seeded!</info>")

    def handle(self):
        tenants = self.tenancy.get_tenants(self.option("tenants"))

        if len(tenants) == 0:
            self.error("No tenants found!")
            exit()

        for tenant in tenants:
            self.info(f"Seeding tenant: {tenant.name}")
            self.warning("=====================START=====================")
            self.seed(tenant)
            self.warning("======================END======================")



import os
from masonite.commands.Command import Command
from masoniteorm.migrations import Migration


class TenancyMigrate(Command):
    """
    Migrates database to all tenants or to a specific tenant.

    tenancy:migrate
        {--tenants=default : List of tenants to migrate}
        {--m|migration=all : Migration's name to be migrated}
        {--f|force : Force migrations without prompt in production}
        {--s|show : Shows the output of SQL for migrations that would be running}
        {--d|directory=databases/migrations : The location of the migration directory}
    """

    def __init__(self, application):
        super().__init__()
        self.app = application
        self.tenancy = self.app.make("multitenancy")

    def migration(self, tenant):
        self.tenancy.setup_connection(tenant)

        migration = Migration(
            command_class=self,
            connection=tenant.database,
            migration_directory=self.option("directory"),
            config_path=None,
            schema=None,
        )

        migration.create_table_if_not_exists()
        if not migration.get_unran_migrations():
            self.info(f"Nothing to migrate for tenant: {tenant.name}!")
            return

        return migration

    def handle(self):
        if os.getenv("APP_ENV") == "production" and not self.option("force"):
            answer = ""
            while answer not in ["y", "n"]:
                answer = input("Do you want to run migrations in PRODUCTION ? (y/n)\n").lower()
            if answer != "y":
                self.info("Migrations cancelled")
                exit(0)

        tenants = self.tenancy.get_tenants(self.option("tenants"))

        if len(tenants) == 0:
            self.error("No tenants found!")
            exit()

        for tenant in tenants:
            migration = self.migration(tenant)
            if migration:
                migration_name = self.option("migration")
                show_output = self.option("show")
                self.info(f"Migrating tenant: {tenant.name}")
                self.warning("=====================START=====================")
                migration.migrate(migration=migration_name, output=show_output)
                self.warning("======================END======================")



from masonite.commands.Command import Command
from masoniteorm.migrations import Migration


class TenancyMigrateStatus(Command):
    """
    Display migration status of all tenants or of a specific tenant.

    tenancy:migrate:status
        {--tenants=default : List of tenants}
        {--d|directory=databases/migrations : The location of the migration directory}
    """

    def __init__(self, application):
        super().__init__()
        self.app = application
        self.tenancy = self.app.make("multitenancy")

    def migration(self, tenant):
        self.tenancy.setup_connection(tenant)

        migration = Migration(
            command_class=self,
            connection=tenant.database,
            migration_directory=self.option("directory"),
            config_path=None,
            schema=None,
        )
        migration.create_table_if_not_exists()
        table = self.table()
        table.set_header_row(["Ran?", "Migration"])
        migrations = []

        for migration_file in migration.get_ran_migrations():
            migrations.append(["<info>Y</info>", f"<comment>{migration_file}</comment>"])

        for migration_file in migration.get_unran_migrations():
            migrations.append(["<error>N</error>", f"<comment>{migration_file}</comment>"])

        table.set_rows(migrations)

        table.render(self.io)

    def handle(self):
        tenants = self.tenancy.get_tenants(self.option("tenants"))

        if len(tenants) == 0:
            self.error("No tenants found!")
            exit()

        for tenant in tenants:
            self.info(f"Status of tenant: {tenant.name}")
            self.warning("=====================START=====================")
            self.migration(tenant)
            self.warning("======================END======================")



from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="masonite-multitenancy",
    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version="0.0.4",
    packages=[
        "multitenancy",
        "multitenancy.commands",
        "multitenancy.config",
        "multitenancy.middlewares",
        "multitenancy.migrations",
        "multitenancy.models",
        "multitenancy.providers",
    ],
    package_dir={"": "src"},
    description="Multitenancy package for Masonite!",
    long_description=long_description,
    long_description_content_type="text/markdown",
    # The project's main homepage.
    url="https://github.com/py-package/masonite-multitenancy",
    # Author details
    author="Yubaraj Shrestha",
    author_email="yubaraj@pypackage.com",
    # Choose your license
    license="MIT license",
    # If your package should include things you specify in your MANIFEST.in file
    # Use this option if your package needs to include files that are not python files
    # like html templates or css files
    include_package_data=True,
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 5 - Production/Stable",
        # Indicate who your project is intended for
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "Environment :: Web Environment",
        # Pick your license as you wish (should match "license" above)
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Internet :: WWW/HTTP :: WSGI",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Topic :: Software Development :: Libraries :: Python Modules",
        # List package on masonite packages website
        "Framework :: Masonite",
    ],
    # What does your project relate to?
    keywords="Masonite, Python, Development",
    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=["masonite>=4.0<5.0"],
    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    # $ pip install your-package[dev,test]
    extras_require={
        "dev": [
            "black",
            "flake8",
            "coverage",
            "pytest",
            "pytest-cov",
            "twine>=1.5.0",
            "wheel",
        ],
    },
    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    package_data={
        # 'templates/index.html': [],
    },
)



"""A MultitenancyProvider Service Provider."""

from masonite.packages import PackageProvider

from ..middlewares.tenant_finder_middleware import TenantFinderMiddleware
from ..commands.TenancyMigrate import TenancyMigrate
from ..commands.TenancyCreate import TenancyCreate
from ..commands.TenancyList import TenancyList
from ..commands.TenancyDelete import TenancyDelete
from ..commands.TenancyMigrateRefresh import TenancyMigrateRefresh
from ..commands.TenancyMigrateRollback import TenancyMigrateRollback
from ..commands.TenancyMigrateReset import TenancyMigrateReset
from ..commands.TenancyMigrateStatus import TenancyMigrateStatus
from ..commands.TenancySeed import TenancySeed
from ..multitenancy import MultiTenancy


class MultitenancyProvider(PackageProvider):
    def configure(self):
        """Register objects into the Service Container."""
        (
            self.root("multitenancy")
            .name("multitenancy")
            .config("config/multitenancy.py", publish=True)
        )

    def register(self):
        super().register()
        self.application.bind("multitenancy", MultiTenancy(self.application))
        self.application.make("middleware").add({"multitenancy": [TenantFinderMiddleware]})
        (
            self.application.make("commands")
            .add(TenancyMigrate(self.application))
            .add(TenancyCreate(self.application))
            .add(TenancyList(self.application))
            .add(TenancyDelete(self.application))
            .add(TenancyMigrateRefresh(self.application))
            .add(TenancyMigrateRollback(self.application))
            .add(TenancyMigrateStatus(self.application))
            .add(TenancyMigrateReset(self.application))
            .add(TenancySeed(self.application))
        )

    def boot(self):
        """Boots services required by the container."""
        pass



from masonite.commands import Command
from ..models.Tenant import Tenant


class TenancyList(Command):
    """
    List all tenants.

    tenancy:list
    """

    def __init__(self, application):
        super().__init__()
        self.app = application

    def handle(self):
        tenants = Tenant.all()

        if len(tenants) == 0:
            self.error("No tenants found!")
            exit()

        self.info("=================Tenants List=================")
        for tenant in tenants:
            print("Tenant: " + tenant.name)
            print("Domain: " + tenant.domain)
            print("Database: " + tenant.database)
            print("----------------------------------------------")
